import six

from .. import errors
from .base import Column, size_by_type


class IntColumn(Column):
    py_types = six.integer_types

    def __init__(self):
        super(IntColumn, self).__init__()

        self.mask = (1 << 8 * size_by_type[self.ch_type]) - 1

    # Chop only bytes that fit current type
    def _prepare(self, value):
        sign = 1 if value > 0 else -1
        return sign * (abs(value) & self.mask)

    def read(self, buf):
        return self._read(buf, self.format)

    def write(self, value, buf):
        self._write(self._prepare(value), buf, self.format)


class UIntColumn(IntColumn):
    def write(self, value, buf):
        if value < 0:
            raise errors.TypeMismatchError(
                'Type mismatch in VALUES section. '
                'Expected {} got {}'.format(self.ch_type, value)
            )

        super(UIntColumn, self).write(value, buf)


class Int8Column(IntColumn):
    ch_type = 'Int8'
    format = '<b'


class Int16Column(IntColumn):
    ch_type = 'Int16'
    format = '<h'


class Int32Column(IntColumn):
    ch_type = 'Int32'
    format = '<i'


class Int64Column(IntColumn):
    ch_type = 'Int64'
    format = '<q'


class UInt8Column(UIntColumn):
    ch_type = 'UInt8'
    format = '<B'


class UInt16Column(UIntColumn):
    ch_type = 'UInt16'
    format = '<H'


class UInt32Column(UIntColumn):
    ch_type = 'UInt32'
    format = '<I'


class UInt64Column(UIntColumn):
    ch_type = 'UInt64'
    format = '<Q'



from enum import Enum

import six

from .. import errors
from .intcolumn import IntColumn


class EnumColumn(IntColumn):
    py_types = (Enum, ) + six.integer_types + six.string_types
    format = '<b'

    def __init__(self, enum_cls):
        self.enum_cls = enum_cls
        super(EnumColumn, self).__init__()

    def read(self, buf):
        value = super(EnumColumn, self).read(buf)
        return self.enum_cls(value).name

    def write(self, value, buf):
        source_value = value.name if isinstance(value, Enum) else value
        enum_cls = self.enum_cls

        # Check real enum value
        try:
            if isinstance(source_value, six.string_types):
                value = enum_cls[source_value].value
            else:
                value = enum_cls(source_value).value

        except (ValueError, KeyError):
            choices = ', '.join(
                "'{}' = {}".format(x.name, x.value) for x in enum_cls
            )
            enum_str = '{}({})'.format(enum_cls.__name__, choices)

            raise errors.LogicalError(
                "Unknown element '{}' for type {}"
                .format(source_value, enum_str)
            )

        super(EnumColumn, self).write(value, buf)


class Enum8Column(EnumColumn):
    ch_type = 'Enum8'
    format = '<b'


class Enum16Column(EnumColumn):
    ch_type = 'Enum16'
    format = '<h'


def create_enum_column(spec):
    if spec.startswith('Enum8'):
        params = spec[6:-1]
        cls = Enum8Column
    else:
        params = spec[7:-1]
        cls = Enum16Column

    d = {}
    for param in params.split(", '"):
        pos = param.rfind("'")
        name = param[:pos].lstrip("'")
        value = int(param[pos + 1:].lstrip(' ='))
        d[name] = value

    return cls(Enum(cls.ch_type, d))



import struct

from .base import Column


class FloatColumn(Column):
    py_types = (float, int)
    format = None
    inf = float('inf')

    def read(self, buf):
        return self._read(buf, self.format)

    def write(self, value, buf):
        try:
            x = struct.pack(self.format, value)
        except OverflowError:
            # Write +/- inf if float overflows
            # Client has it's behaviour now.
            value = self.inf if value > 0 else -self.inf
            x = struct.pack(self.format, value)

        buf.write(x)


class Float32(FloatColumn):
    ch_type = 'Float32'
    format = '<f'


class Float64(FloatColumn):
    ch_type = 'Float64'
    format = '<d'



from .reader import read_varint, read_binary_uint8, read_binary_int32
from .writer import write_varint, write_binary_uint8, write_binary_int32


class BlockInfo(object):
    is_overflows = False
    bucket_num = -1

    def write(self, buf):
        # Set of pairs (`FIELD_NUM`, value) in binary form. Then 0.
        write_varint(1, buf)
        write_binary_uint8(self.is_overflows, buf)

        write_varint(2, buf)
        write_binary_int32(self.bucket_num, buf)

        write_varint(0, buf)

    def read(self, buf):
        while True:
            field_num = read_varint(buf)
            if not field_num:
                break

            if field_num == 1:
                self.is_overflows = bool(read_binary_uint8(buf))

            elif field_num == 2:
                self.bucket_num = read_binary_int32(buf)


class Block(object):
    def __init__(self, columns_with_types=None, data=None, info=None):
        self.columns_with_types = columns_with_types or []
        self.data = data or []

        # Transpose data rows->columns
        if data:
            # Guessing about whole data format by first row.
            # Converting dicts to lists.
            columns = [x[0] for x in self.columns_with_types]
            if isinstance(data[0], dict):
                data = [tuple(row[c] for c in columns) for row in data]

            self.data = list(zip(*data))

        self.info = info or BlockInfo()

        super(Block, self).__init__()

    @classmethod
    def check_data_sanity(cls, columns_with_types, data):
        if not data:
            return

        # Check each row has the same length.
        row_len = len(data[0])
        for row in data:
            if len(row) != row_len:
                raise ValueError('Different rows length')

        columns_expected = len(columns_with_types)
        if columns_expected != row_len:
            raise ValueError(
                'Expected {} columns, got {}'.format(columns_expected, row_len)
            )

    @property
    def columns(self):
        return len(self.data)

    @property
    def rows(self):
        return len(self.data[0]) if self.columns else 0



import struct

from six import PY3


if PY3:
    def _byte(b):
        return bytes((b, ))
else:
    def _byte(b):
        return chr(b)


def write_binary_str(text, buf):
    text = text.encode('utf-8')
    write_varint(len(text), buf)
    buf.write(text)


def write_binary_str_fixed_len(text, buf, length):
    text = text.encode('utf-8')
    diff = length - len(text)
    if diff > 0:
        text += _byte(0) * diff
    elif diff < 0:
        raise ValueError
    buf.write(text)


def write_varint(number, buf):
    """
    Writes integer of variable length using LEB128.
    """
    while True:
        towrite = number & 0x7f
        number >>= 7
        if number:
            buf.write(_byte(towrite | 0x80))
        else:
            buf.write(_byte(towrite))
            break


def write_binary_int(number, buf, fmt):
    """
    Writes int from buffer with provided format.
    """
    fmt = '<' + fmt
    buf.write(struct.pack(fmt, number))


def write_binary_int8(number, buf):
    write_binary_int(number, buf, 'b')


def write_binary_int16(number, buf):
    write_binary_int(number, buf, 'h')


def write_binary_int32(number, buf):
    write_binary_int(number, buf, 'i')


def write_binary_int64(number, buf):
    write_binary_int(number, buf, 'q')


def write_binary_uint8(number, buf):
    write_binary_int(number, buf, 'B')


def write_binary_uint16(number, buf):
    write_binary_int(number, buf, 'H')


def write_binary_uint32(number, buf):
    write_binary_int(number, buf, 'I')


def write_binary_uint64(number, buf):
    write_binary_int(number, buf, 'Q')


MAX_UINT64 = (1 << 64) - 1


def write_binary_uint128(number, buf):
    fmt = '<QQ'
    packed = struct.pack(fmt, (number >> 64) & MAX_UINT64, number & MAX_UINT64)
    buf.write(packed)



from __future__ import absolute_import
from io import BytesIO

import quicklz

from .base import BaseCompressor, BaseDecompressor
from ..protocol import CompressionMethod
from ..reader import read_binary_uint32, read_binary_uint8
from ..writer import write_binary_uint32, write_binary_uint8


class Compressor(BaseCompressor):
    method = CompressionMethod.QUICK_LZ

    def get_compressed_data(self, extra_header_size):
        compressed = BytesIO(quicklz.compress(self.data.getvalue()))

        block_size = read_binary_uint8(compressed)
        rv = BytesIO()
        write_binary_uint8(block_size & 3, rv)
        rv.write(compressed.read())
        return rv.getvalue()


class Decompressor(BaseDecompressor):
    method = CompressionMethod.QUICK_LZ

    def _get_compressed_size(self, stream):
        block_size = read_binary_uint8(stream)
        big_block = (block_size & 2) == 2

        if big_block:
            read = read_binary_uint32
            write = write_binary_uint32
            s = 4

        else:
            read = read_binary_uint8
            write = write_binary_uint32
            s = 1

        r = read(stream)
        size = r & (0xffffffff >> ((4 - s) * 8))

        header = BytesIO()
        write_binary_uint8(block_size, header)
        write(r, header)

        return size - (1 + s), header.getvalue()

    def get_decompressed_data(self, method_byte, compressed_hash,
                              extra_header_size):
        remain_size, header_bytes = self._get_compressed_size(self.stream)

        compressed = BytesIO(header_bytes)
        compressed.write(self.stream.read(remain_size))

        self.check_hash(compressed.getvalue(), compressed_hash)

        return quicklz.decompress(compressed)



import six

from .. import errors
from ..reader import read_binary_str, read_binary_str_fixed_len
from ..writer import write_binary_str, write_binary_str_fixed_len
from .base import Column


class String(Column):
    ch_type = 'String'
    py_types = six.string_types

    def read(self, buf):
        return read_binary_str(buf)

    def write(self, value, buf):
        write_binary_str(value, buf)


class FixedString(Column):
    ch_type = 'FixedString'
    py_types = six.string_types

    def __init__(self, length):
        self.length = length
        super(FixedString, self).__init__()

    def read(self, buf):
        return read_binary_str_fixed_len(buf, self.length).strip('\x00')

    def write(self, value, buf):
        try:
            write_binary_str_fixed_len(value, buf, self.length)
        except ValueError:
            raise errors.TooLargeStringSize()



import six

from .base import Column
from .intcolumn import UInt64Column

if six.PY3:
    from queue import Queue
else:
    from Queue import Queue


class ArrayColumn(Column):
    """
    Nested arrays written in flatten form after information about their
    sizes (offsets really).
    One element of array of arrays can be represented as tree:
    (0 depth)          [[3, 4], [5, 6]]
                      /               \
    (1 depth)      [3, 4]           [5, 6]
                   /    \           /    \
    (leaf)        3      4         5      6

    Offsets (sizes) written in breadth-first search order. In example above
    following sequence of offset will be written: 4 -> 2 -> 4
    1) size of whole array: 4
    2) size of array 1 in depth=1: 2
    3) size of array 2 plus size of all array before in depth=1: 2 + 2 = 4

    After sizes info comes flatten data: 3 -> 4 -> 5 -> 6
    """
    py_types = (list, tuple)

    def __init__(self, nested_column):
        self.size_column = UInt64Column()
        self.nested_column = nested_column
        self._write_depth_0_size = True
        super(ArrayColumn, self).__init__()

    def write_data(self, data, buf):
        # Column of Array(T) is stored in "compact" format and passed to server
        # wrapped into another Array without size of wrapper array.
        self.nested_column = ArrayColumn(self.nested_column)
        self._write_depth_0_size = False
        super(ArrayColumn, self).write_data((data, ), buf)

    def read_data(self, rows, buf):
        self.nested_column = ArrayColumn(self.nested_column)
        return self._read_data(rows, buf)

    def _write_sizes(self, value, buf):
        q = Queue()
        q.put((self, value, 0))

        cur_depth = 0
        offset = 0
        while not q.empty():
            column, value, depth = q.get_nowait()

            if cur_depth != depth:
                cur_depth = depth
                offset = 0

            offset += len(value)
            if (cur_depth == 0 and self._write_depth_0_size) or cur_depth > 0:
                self.size_column.write(offset, buf)

            nested_column = column.nested_column
            if isinstance(nested_column, ArrayColumn):
                for x in value:
                    q.put((nested_column, x, cur_depth + 1))

    def _write_data(self, value, buf):
        if isinstance(self.nested_column, ArrayColumn):
            for x in value:
                self.nested_column._write_data(x, buf)
        else:
            for x in value:
                self.nested_column.write(x, buf)

    def write(self, value, buf):
        self._write_sizes(value, buf)
        self._write_data(value, buf)

    def _read_data(self, size, buf):
        q = Queue()
        q.put((self, size, 0))

        data = []
        slices_series = []

        cur_depth = 0
        prev_offset = 0
        slices = []
        # Read and store info about slices.
        while not q.empty():
            column, size, depth = q.get_nowait()

            if cur_depth != depth:
                cur_depth = depth
                prev_offset = 0
                slices_series.append(slices)
                slices = []

            nested_column = column.nested_column
            if isinstance(nested_column, ArrayColumn):
                for _i in range(size):
                    offset = self.size_column.read(buf)
                    q.put((nested_column, offset - prev_offset, cur_depth + 1))
                    slices.append((prev_offset, offset))
                    prev_offset = offset

            # Read data
            else:
                data.extend(nested_column.read(buf) for _i in range(size))

        # Build nested tuple structure.
        for slices in reversed(slices_series):
            nested_data = []
            for slice_from, slice_to in slices:
                nested_data.append(tuple(data[slice_from:slice_to]))
            data = nested_data

        return tuple(data)


def create_array_column(spec, column_by_spec_getter):
    inner = spec[6:-1]
    return ArrayColumn(column_by_spec_getter(inner))



from io import BytesIO

from clickhouse_driver.util.cityhash import CityHash128

from .native import BlockOutputStream, BlockInputStream
from ..reader import read_binary_uint8, read_binary_uint128
from ..writer import write_binary_uint8, write_binary_uint128
from ..compression import get_decompressor_cls


class CompressedBlockOutputStream(BlockOutputStream):
    def __init__(self, compressor_cls, compress_block_size, fout,
                 server_revision):
        self.compressor_cls = compressor_cls
        self.compress_block_size = compress_block_size
        self.raw_fout = fout

        self.compressor = self.compressor_cls()
        super(CompressedBlockOutputStream, self).__init__(self.compressor,
                                                          server_revision)

    def reset(self):
        self.compressor = self.compressor_cls()
        self.fout = self.compressor

    def get_compressed_hash(self, data):
        return CityHash128(data)

    def finalize(self):
        compressed = self.get_compressed()
        compressed_size = len(compressed)

        compressed_hash = self.get_compressed_hash(compressed)
        write_binary_uint128(compressed_hash, self.raw_fout)

        block_size = self.compress_block_size

        i = 0
        while i < compressed_size:
            self.raw_fout.write(compressed[i:i + block_size])
            i += block_size

        self.raw_fout.flush()

    def get_compressed(self):
        compressed = BytesIO()

        if self.compressor.method_byte is not None:
            write_binary_uint8(self.compressor.method_byte, compressed)
            extra_header_size = 1  # method
        else:
            extra_header_size = 0

        data = self.compressor.get_compressed_data(extra_header_size)
        compressed.write(data)

        return compressed.getvalue()


class CompressedBlockReader(object):
    def __init__(self, read_block):
        self.read_block = read_block
        self.block = None

        super(CompressedBlockReader, self).__init__()

    def read(self, n=-1):
        if not self.block:
            self.block = BytesIO(self.read_block())

        rv = self.block.read(n)
        read = len(rv)

        if n != -1 and read != n:
            self.block = BytesIO(self.read_block())
            unread = n - read
            rv += self.block.read(unread)

        return rv


class CompressedBlockInputStream(BlockInputStream):
    def __init__(self, fin, server_revision):
        self.raw_fin = fin
        fin = CompressedBlockReader(self.read_block)
        super(CompressedBlockInputStream, self).__init__(fin, server_revision)

    def get_compressed_hash(self, data):
        return CityHash128(data)

    def read_block(self):
        compressed_hash = read_binary_uint128(self.raw_fin)
        method_byte = read_binary_uint8(self.raw_fin)

        decompressor_cls = get_decompressor_cls(method_byte)
        decompressor = decompressor_cls(self.raw_fin)

        if decompressor.method_byte is not None:
            extra_header_size = 1  # method
        else:
            extra_header_size = 0

        return decompressor.get_decompressed_data(
            method_byte, compressed_hash, extra_header_size
        )



from contextlib import contextmanager
import logging
import socket

from .block import Block
from .blockstreamprofileinfo import BlockStreamProfileInfo
from .clientinfo import ClientInfo
from . import defines
from . import errors
from .progress import Progress
from .protocol import Compression, ClientPacketTypes, ServerPacketTypes
from .queryprocessingstage import QueryProcessingStage
from .reader import read_varint, read_binary_str
from .readhelpers import read_exception
from .compression import get_compressor_cls
from .writer import write_varint, write_binary_str


logger = logging.getLogger(__name__)


class Packet(object):
    def __init__(self):
        self.type = None
        self.block = None
        self.exception = None
        self.progress = None
        self.profile_info = None

        super(Packet, self).__init__()


class ServerInfo(object):
    def __init__(self, name, version_major, version_minor, revision, timezone):
        self.name = name
        self.version_major = version_major
        self.version_minor = version_minor
        self.revision = revision
        self.timezone = timezone

        super(ServerInfo, self).__init__()


class Connection(object):
    def __init__(
            self, host, port=defines.DEFAULT_PORT,
            database='default', user='default', password='',
            client_name=defines.CLIENT_NAME,
            connect_timeout=defines.DBMS_DEFAULT_CONNECT_TIMEOUT_SEC,
            send_receive_timeout=defines.DBMS_DEFAULT_TIMEOUT_SEC,
            sync_request_timeout=defines.DBMS_DEFAULT_SYNC_REQUEST_TIMEOUT_SEC,
            compress_block_size=defines.DEFAULT_COMPRESS_BLOCK_SIZE,
            compression=False
    ):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password
        self.client_name = defines.DBMS_NAME + ' ' + client_name
        self.connect_timeout = connect_timeout
        self.send_receive_timeout = send_receive_timeout
        self.sync_request_timeout = sync_request_timeout

        # Use LZ4 compression by default.
        if compression is True:
            compression = 'lz4'

        if compression is False:
            self.compression = Compression.DISABLED
            self.compressor_cls = None
            self.compress_block_size = None
        else:
            self.compression = Compression.ENABLED
            self.compressor_cls = get_compressor_cls(compression)
            self.compress_block_size = compress_block_size

        self.socket = None
        self.fin = None
        self.fout = None

        self.connected = False

        self.server_info = None

        # Block writer/reader
        self.block_in = None
        self.block_out = None

        super(Connection, self).__init__()

    def get_description(self):
        return '{}:{}'.format(self.host, self.port)

    def force_connect(self):

        if not self.connected:
            self.connect()

        elif not self.ping():
            logger.info('Connection was closed, reconnecting.')
            self.connect()

    def connect(self):
        try:
            if self.connected:
                self.disconnect()

            logger.info(
                'Connecting. Database: %s. User: %s', self.database, self.user
            )

            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.settimeout(self.connect_timeout)
            self.socket.connect((self.host, self.port))
            self.connected = True
            self.socket.settimeout(self.send_receive_timeout)

            # performance tweak
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            self.fin = self.socket.makefile('rb')
            self.fout = self.socket.makefile('wb')

            self.send_hello()
            self.receive_hello()

            self.block_in = self.get_block_in_stream()
            self.block_out = self.get_block_out_stream()

        except socket.timeout as e:
            self.disconnect()
            raise errors.SocketTimeoutError(
                '{} ({})'.format(e.message, self.get_description())
            )

        except socket.error as e:
            self.disconnect()
            raise errors.NetworkError(
                '{} ({})'.format(e.strerror, self.get_description())
            )

    def reset_state(self):
        self.socket = None
        self.fin = None
        self.fout = None

        self.connected = False

        self.server_info = None

        self.block_in = None
        self.block_out = None

    def disconnect(self):
        if self.connected:
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()

        self.reset_state()

    def send_hello(self):
        write_varint(ClientPacketTypes.HELLO, self.fout)
        write_binary_str(self.client_name, self.fout)
        write_varint(defines.DBMS_VERSION_MAJOR, self.fout)
        write_varint(defines.DBMS_VERSION_MINOR, self.fout)
        write_varint(defines.CLIENT_VERSION, self.fout)
        write_binary_str(self.database, self.fout)
        write_binary_str(self.user, self.fout)
        write_binary_str(self.password, self.fout)

        self.fout.flush()

    def receive_hello(self):
        packet_type = read_varint(self.fin)

        if packet_type == ServerPacketTypes.HELLO:
            server_name = read_binary_str(self.fin)
            server_version_major = read_varint(self.fin)
            server_version_minor = read_varint(self.fin)
            server_revision = read_varint(self.fin)

            server_timezone = None
            if server_revision >= \
                    defines.DBMS_MIN_REVISION_WITH_SERVER_TIMEZONE:
                server_timezone = read_binary_str(self.fin)

            self.server_info = ServerInfo(
                server_name, server_version_major, server_version_minor,
                server_revision, server_timezone
            )

            logger.info(
                'Connected to %s server version %s.%s.%s', server_name,
                server_version_major, server_version_minor, server_revision
            )

        elif packet_type == ServerPacketTypes.EXCEPTION:
            raise self.receive_exception()

        else:
            self.disconnect()
            message = self.unexpected_packet_message('Hello or Exception',
                                                     packet_type)
            raise errors.UnexpectedPacketFromServerError(message)

    def ping(self):
        timeout = self.sync_request_timeout

        with self.timeout_setter(timeout):
            try:
                write_varint(ClientPacketTypes.PING, self.fout)
                self.fout.flush()

                packet_type = read_varint(self.fin)
                while packet_type == ServerPacketTypes.PROGRESS:
                    self.receive_progress()
                    packet_type = read_varint(self.fin)

                if packet_type != ServerPacketTypes.PONG:
                    msg = self.unexpected_packet_message('Pong', packet_type)
                    raise errors.UnexpectedPacketFromServerError(msg)

            except Exception as e:
                logger.exception(e)
                return False

        return True

    def receive_packet(self):
        packet = Packet()

        packet.type = packet_type = read_varint(self.fin)

        if packet_type == ServerPacketTypes.DATA:
            packet.block = self.receive_data()

        elif packet_type == ServerPacketTypes.EXCEPTION:
            packet.exception = self.receive_exception()

        elif packet.type == ServerPacketTypes.PROGRESS:
            packet.progress = self.receive_progress()

        elif packet.type == ServerPacketTypes.PROFILE_INFO:
            packet.profile_info = self.receive_profile_info()

        elif packet_type == ServerPacketTypes.TOTALS:
            packet.block = self.receive_data()

        elif packet_type == ServerPacketTypes.EXTREMES:
            packet.block = self.receive_data()

        elif packet_type == ServerPacketTypes.END_OF_STREAM:
            pass

        else:
            self.disconnect()
            raise errors.UnknownPacketFromServerError(
                'Unknown packet {} from server {}'.format(
                    packet_type, self.get_description()
                )
            )

        return packet

    def get_block_in_stream(self):
        revision = self.server_info.revision

        if self.compression:
            from .streams.compressed import CompressedBlockInputStream

            return CompressedBlockInputStream(self.fin, revision)
        else:
            from .streams.native import BlockInputStream

            return BlockInputStream(self.fin, revision)

    def get_block_out_stream(self):
        revision = self.server_info.revision

        if self.compression:
            from .streams.compressed import CompressedBlockOutputStream

            return CompressedBlockOutputStream(
                self.compressor_cls, self.compress_block_size,
                self.fout, revision
            )
        else:
            from .streams.native import BlockOutputStream

            return BlockOutputStream(self.fout, revision)

    def receive_data(self):
        revision = self.server_info.revision

        if revision >= defines.DBMS_MIN_REVISION_WITH_TEMPORARY_TABLES:
            read_binary_str(self.fin)

        block = self.block_in.read()
        self.block_in.reset()
        return block

    def receive_exception(self):
        return read_exception(self.fin)

    def receive_progress(self):
        progress = Progress()
        progress.read(self.server_info.revision, self.fin)
        return progress

    def receive_profile_info(self):
        profile_info = BlockStreamProfileInfo()
        profile_info.read(self.fin)
        return profile_info

    def send_data(self, block, table_name=''):
        write_varint(ClientPacketTypes.DATA, self.fout)

        revision = self.server_info.revision
        if revision >= defines.DBMS_MIN_REVISION_WITH_TEMPORARY_TABLES:
            write_binary_str(table_name, self.fout)

        self.block_out.write(block)
        self.block_out.reset()

    def send_query(self, query):
        if not self.connected:
            self.connect()

        write_varint(ClientPacketTypes.QUERY, self.fout)

        query_id = ''
        write_binary_str(query_id, self.fout)

        revision = self.server_info.revision
        if revision >= defines.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
            client_info = ClientInfo(self.client_name)
            client_info.query_kind = ClientInfo.QueryKind.INITIAL_QUERY

            client_info.write(revision, self.fout)

        write_binary_str('', self.fout)  # query settings

        write_varint(QueryProcessingStage.COMPLETE, self.fout)
        write_varint(self.compression, self.fout)

        write_binary_str(query, self.fout)

        logger.info('Query: %s', query)

        self.fout.flush()

    def send_cancel(self):
        write_varint(ClientPacketTypes.CANCEL, self.fout)

        self.fout.flush()

    def send_external_tables(self, tables):
        for table in tables:
            block = Block(table['structure'], table['data'])
            self.send_data(block, table_name=table['name'])

        # Empty block, end of data transfer.
        self.send_data(Block())

    @contextmanager
    def timeout_setter(self, new_timeout):
        old_timeout = self.socket.gettimeout()
        self.socket.settimeout(new_timeout)

        yield

        self.socket.settimeout(old_timeout)

    def unexpected_packet_message(self, expected, packet_type):
        packet_type = ServerPacketTypes.to_str(packet_type)

        return (
            'Unexpected packet from server {} (expected {}, got {})'
            .format(self.get_description(), expected, packet_type)
        )



from calendar import timegm
from datetime import datetime

from ..util.tzinfo import tzutc
from .base import Column


class DateTimeColumn(Column):
    ch_type = 'DateTime'
    # TODO: string
    py_types = (datetime, )
    format = '<I'

    utc = tzutc()

    def read(self, buf):
        x = self._read(buf, self.format)
        return datetime.fromtimestamp(x, tz=self.utc).replace(tzinfo=None)

    def write(self, value, buf):
        x = int(timegm(value.timetuple()))
        self._write(x, buf, self.format)



import struct


size_by_type = {
    'Date': 2,
    'DateTime': 4,
    'Int8': 1,
    'UInt8': 1,
    'Int16': 2,
    'UInt16': 2,
    'Int32': 4,
    'UInt32': 4,
    'Int64': 8,
    'UInt64': 8,
    'Float32': 4,
    'Float64': 8,
    'Enum8': 1,
    'Enum16': 2
}


class Column(object):
    ch_type = None
    py_types = None

    @property
    def size(self):
        return size_by_type[self.ch_type]

    def read(self, buf):
        raise NotImplementedError

    def write(self, value, buf):
        raise NotImplementedError

    def _read(self, buf, fmt):
        return struct.unpack(fmt, buf.read(self.size))[0]

    def _write(self, x, buf, fmt):
        return buf.write(struct.pack(fmt, x))

    def write_data(self, data, buf):
        for x in data:
            if not isinstance(x, self.py_types):
                raise TypeError(x)

            self.write(x, buf)

    def read_data(self, rows, buf):
        return tuple(self.read(buf) for _i in range(rows))




import six


class ErrorCodes(object):
    NO_SUCH_COLUMN_IN_TABLE = 16
    CHECKSUM_DOESNT_MATCH = 40
    LOGICAL_ERROR = 49
    UNKNOWN_TYPE = 50
    TYPE_MISMATCH = 53
    UNKNOWN_COMPRESSION_METHOD = 89
    TOO_LARGE_STRING_SIZE = 131
    NETWORK_ERROR = 210
    SOCKET_TIMEOUT = 209
    SERVER_REVISION_IS_TOO_OLD = 197
    UNEXPECTED_PACKET_FROM_SERVER = 102
    UNKNOWN_PACKET_FROM_SERVER = 100


if six.PY3:
    class Error(Exception):
        code = None

        def __init__(self, message=None):
            self.message = message

        def __str__(self):
            message = ' ' + self.message if self.message is not None else ''
            return 'Code: {}.{}'.format(self.code, message)

    class ServerException(Error):
        def __init__(self, message, code, nested=None):
            self.message = message
            self.code = code
            self.nested = nested

        def __str__(self):
            nested = '\nNested: {}'.format(self.nested) if self.nested else ''
            return 'Code: {}.{}\n{}'.format(self.code, nested, self.message)

else:
    class Error(Exception):
        code = None

        def __init__(self, message=None):
            self.message = message

        def __unicode__(self):
            message = ' ' + self.message if self.message is not None else ''
            return 'Code: {}.{}'.format(self.code, message)

        def __str__(self):
            return six.text_type(self).encode('utf-8')

    class ServerException(Error):
        def __init__(self, message, code, nested=None):
            self.message = message
            self.code = code
            self.nested = nested

        def __unicode__(self):
            nested = '\nNested: {}'.format(self.nested) if self.nested else ''
            return 'Code: {}.{}\n{}'.format(self.code, nested, self.message)

        def __str__(self):
            return six.text_type(self).encode('utf-8')


class LogicalError(Error):
    code = ErrorCodes.LOGICAL_ERROR


class UnknownTypeError(Error):
    code = ErrorCodes.UNKNOWN_TYPE


class ChecksumDoesntMatchError(Error):
    code = ErrorCodes.CHECKSUM_DOESNT_MATCH


class TypeMismatchError(Error):
    code = ErrorCodes.TYPE_MISMATCH


class UnknownCompressionMethod(Error):
    code = ErrorCodes.UNKNOWN_COMPRESSION_METHOD


class TooLargeStringSize(Error):
    code = ErrorCodes.TOO_LARGE_STRING_SIZE


class NetworkError(Error):
    code = ErrorCodes.NETWORK_ERROR


class SocketTimeoutError(Error):
    code = ErrorCodes.SOCKET_TIMEOUT


class UnexpectedPacketFromServerError(Error):
    code = ErrorCodes.UNEXPECTED_PACKET_FROM_SERVER


class UnknownPacketFromServerError(Error):
    code = ErrorCodes.UNKNOWN_PACKET_FROM_SERVER



from .. import errors
from .arraycolumn import create_array_column
from .datecolumn import DateColumn
from .datetimecolumn import DateTimeColumn
from .enumcolumn import create_enum_column
from .floatcolumn import Float32, Float64
from .intcolumn import (
    Int8Column, Int16Column, Int32Column, Int64Column,
    UInt8Column, UInt16Column, UInt32Column, UInt64Column
)
from .stringcolumn import String, FixedString


column_by_type = {c.ch_type: c for c in [
    DateColumn, DateTimeColumn, String, Float32, Float64,
    Int8Column, Int16Column, Int32Column, Int64Column,
    UInt8Column, UInt16Column, UInt32Column, UInt64Column
]}


def get_column_by_spec(spec):
    if spec.startswith('FixedString'):
        length = int(spec[12:-1])
        return FixedString(length)

    elif spec.startswith('Enum'):
        return create_enum_column(spec)

    elif spec.startswith('Array'):
        return create_array_column(spec, get_column_by_spec)

    else:
        try:
            cls = column_by_type[spec]
            return cls()

        except KeyError as e:
            raise errors.UnknownTypeError('Unknown type {}'.format(e.args[0]))


def read_column(column_spec, rows, buf):
    column = get_column_by_spec(column_spec)
    return column.read_data(rows, buf)


def write_column(column_spec, data, buf):
    column = get_column_by_spec(column_spec)

    try:
        column.write_data(data, buf)

    except TypeError as e:
        raise errors.TypeMismatchError(
            'Type mismatch in VALUES section. '
            'Expected {} got {}'.format(column_spec, type(e.args[0]))
        )




class ClientPacketTypes(object):
    """
    Packet types that client transmits
    """
    # Name, version, revision, default DB
    HELLO = 0

    # Query id, query settings, stage up to which the query must be executed,
    # whether the compression must be used, query text
    # (without data for INSERTs).
    QUERY = 1

    # A block of data (compressed or not).
    DATA = 2

    # Cancel the query execution.
    CANCEL = 3

    # Check that connection to the server is alive.
    PING = 4

    # Check status of tables on the server.
    TABLES_STATUS_REQUEST = 5

    _types_str = [
        'Hello', 'Query', 'Data', 'Cancel', 'Ping', 'TablesStatusRequest'
    ]

    @classmethod
    def to_str(cls, packet):
        return 'Unknown packet' if packet < 6 else cls._types_str[packet]


class ServerPacketTypes(object):
    """
    Packet types that server transmits.
    """
    # Name, version, revision.
    HELLO = 0

    # A block of data (compressed or not).
    DATA = 1

    # The exception during query execution.
    EXCEPTION = 2

    # Query execution progress: rows read, bytes read.
    PROGRESS = 3

    # Ping response
    PONG = 4

    # All packets were transmitted
    END_OF_STREAM = 5

    # Packet with profiling info.
    PROFILE_INFO = 6

    # A block with totals (compressed or not).
    TOTALS = 7

    # A block with minimums and maximums (compressed or not).
    EXTREMES = 8

    # A response to TablesStatus request.
    TABLES_STATUS_RESPONSE = 9

    _types_str = [
        'Hello', 'Data', 'Exception', 'Progress', 'Pong', 'EndOfStream',
        'ProfileInfo', 'Totals', 'Extremes', 'TablesStatusResponse'
    ]

    @classmethod
    def to_str(cls, packet):
        return 'Unknown packet' if packet < 10 else cls._types_str[packet]


class Compression(object):
    DISABLED = 0
    ENABLED = 1


class CompressionMethod(object):
    QUICK_LZ = 0
    LZ4 = 1
    LZ4HC = 2
    ZSTD = 3


class CompressionMethodByte(object):
    LZ4 = 0x82
    ZSTD = 0x90



from calendar import timegm
from datetime import date

from .base import Column


class DateColumn(Column):
    ch_type = 'Date'
    # TODO: string
    py_types = (date, )
    format = '<H'

    offset = 24 * 3600

    def read(self, buf):
        x = self._read(buf, self.format)
        return date.fromtimestamp(x * self.offset)

    def write(self, value, buf):
        x = timegm(value.timetuple()) // self.offset
        self._write(x, buf, self.format)



from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='clickhouse-driver',
    version='0.0.4',

    description='Python driver with native interface for ClickHouse',
    long_description=long_description,

    url='https://github.com/mymarilyn/clickhouse-driver',

    author='Konstantin Lebedev',
    author_email='kostyan.lebedev@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',


        'Environment :: Console',


        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',


        'License :: OSI Approved :: MIT License',


        'Operating System :: OS Independent',


        'Programming Language :: SQL',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',


        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],

    keywords='ClickHouse db database cloud analytics',

    packages=[
        p.replace('src', 'clickhouse_driver')
        for p in find_packages(exclude=['tests'])
        if p.startswith('src')
    ],
    package_dir={
        'clickhouse_driver': 'src',
    },
    install_requires=[
        'six',
        'enum34'
    ],
    extras_require={
        'quicklz': ['pyquicklz', 'clickhouse-cityhash==1.0.2'],
        'lz4': ['lz4', 'clickhouse-cityhash==1.0.2'],
        'zstd': ['zstd', 'clickhouse-cityhash==1.0.2']
    },
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'pyquicklz', 'lz4', 'zstd',
        'clickhouse-cityhash==1.0.2'
    ],
)



from .block import Block
from .connection import Connection
from . import errors
from .protocol import ServerPacketTypes


class Client(object):
    def __init__(self, *args, **kwargs):
        self.connection = Connection(*args, **kwargs)
        super(Client, self).__init__()

    def disconnect(self):
        self.connection.disconnect()

    def receive_result(self, with_column_types=False):
        data, columns_with_types = [], []

        while True:
            block = self.receive_block()
            if not block:
                break

            if block is True:
                continue

            # Header block contains no rows. Pick columns from it.
            if block.rows:
                data.extend(block.data)
            elif not columns_with_types:
                columns_with_types = block.columns_with_types

        if with_column_types:
            return data, columns_with_types
        else:
            return data

    def receive_block(self):
        packet = self.connection.receive_packet()

        if packet.type == ServerPacketTypes.EXCEPTION:
            return False

        elif packet.type == ServerPacketTypes.END_OF_STREAM:
            return False

        elif packet.type == ServerPacketTypes.DATA:
            return packet.block

        else:
            return True

    def execute(self, query, params=None, with_column_types=False,
                external_tables=None):
        self.connection.force_connect()

        try:
            is_insert = params is not None
            if is_insert:
                return self.process_insert_query(
                    query, params, external_tables=external_tables
                )
            else:
                return self.process_ordinary_query(
                    query, with_column_types=with_column_types,
                    external_tables=external_tables
                )

        except Exception:
            self.connection.disconnect()
            raise

    def process_ordinary_query(self, query, with_column_types=False,
                               external_tables=None):
        self.connection.send_query(query)
        self.connection.send_external_tables(external_tables or [])
        return self.receive_result(with_column_types=with_column_types)

    def process_insert_query(self, query_without_data, data,
                             external_tables=None):
        self.connection.send_query(query_without_data)
        self.connection.send_external_tables(external_tables or [])

        sample_block = self.receive_sample_block()
        if sample_block:
            self.send_data(sample_block, data)
            packet = self.connection.receive_packet()
            if packet.exception:
                raise packet.exception

    def receive_sample_block(self):
        packet = self.connection.receive_packet()

        if packet.type == ServerPacketTypes.DATA:
            return packet.block

        elif packet.type == ServerPacketTypes.EXCEPTION:
            raise packet.exception

        else:
            message = self.connection.unexpected_packet_message('Data',
                                                                packet.type)
            raise errors.UnexpectedPacketFromServerError(message)

    def send_data(self, sample_block, data):
        Block.check_data_sanity(sample_block.columns_with_types, data)

        block = Block(sample_block.columns_with_types, data)
        self.connection.send_data(block)

        # Empty block means end of data.
        self.connection.send_data(Block())
